#include <stdio.h>
#include <stdlib.h>
#include <math.h>

typedef struct aleaTable_t {
  int nbLignes;
  int nbColonnes;
  int *sommesLignes;
  int *sommesColonnes;
  int **valeurs;} aleaTable_t;


/************************************************************************/
/*                                                                      */
/*Procedure de creation d'une table de contingence a partir d'un fichier*/
/*                                                                      */
/************************************************************************/

void lectureFichier(char *nomFichier,aleaTable_t *aleaTable) {
  int i;
  int *valeurRetour;
  FILE *fichier;

  /*premiere lecture pour compter le nombre de lignes et de colonnes*/
  fichier=fopen(nomFichier,"r");
  valeurRetour=(int *)fscanf(fichier,"%d",&(aleaTable->nbLignes));
  aleaTable->sommesLignes=(int *)malloc(sizeof(int)*aleaTable->nbLignes);
  for(i=0;i<aleaTable->nbLignes;i++) {
     valeurRetour=(int *)fscanf(fichier,"%d",&(aleaTable->sommesLignes[i]));
  }

   valeurRetour=(int *)fscanf(fichier,"%d",&(aleaTable->nbColonnes));
  aleaTable->sommesColonnes=(int *)malloc(sizeof(int)*aleaTable->nbColonnes);
  for(i=0;i<aleaTable->nbColonnes;i++) {
     valeurRetour=(int *)fscanf(fichier,"%d",&(aleaTable->sommesColonnes[i]));
  }
  
  aleaTable->valeurs=(int **)malloc(sizeof(int *)*aleaTable->nbLignes);
  for(i=0;i<aleaTable->nbLignes;i++) {
    aleaTable->valeurs[i]=(int *)malloc(sizeof(int)*aleaTable->nbColonnes);
  }

  
  /*on verifie que les deux sommes sont bien egales*/
  int somme1, somme2;
  somme1=0;
  for (i=0;i<aleaTable->nbLignes;i++) {
    somme1+=aleaTable->sommesLignes[i];
  }
  somme2=0;
  for (i=0;i<aleaTable->nbColonnes;i++) {
    somme2+=aleaTable->sommesColonnes[i];
  }
  if (somme1 != somme2) {
    printf("Table de contingence mal definie\n");
    exit(0);
  }
  aleaTable->valeurs=(int **)malloc(sizeof(int *)*aleaTable->nbLignes);
  for (i=0;i<aleaTable->nbLignes;i++) {
    aleaTable->valeurs[i]=(int *)malloc(sizeof(int)*aleaTable->nbColonnes);
  } 
}

  
/**************************************************/
/*                                                */
/*Procedure de creation d'une table de contingence*/
/*                                                */
/**************************************************/

void aleaTable1(int nbLignes,int nbColonnes,aleaTable_t *aleaTable) {
  int i;
  
  aleaTable->nbLignes = nbLignes;
  aleaTable->nbColonnes = nbColonnes;
  
  aleaTable->valeurs=(int **)malloc(sizeof(int *)*nbLignes);
  for (i=0;i<nbLignes;i++) {
    aleaTable->valeurs[i]=(int *)malloc(sizeof(int)*nbColonnes);
  }
}


/**************************************************/
/*                                                */
/*Procedure de creation d'une table de contingence*/
/*                                                */
/**************************************************/

void aleaTable2(int nbLignes,int nbColonnes,int *sommesLignes,int *sommesColonnes,
		aleaTable_t *aleaTable) {
  int i;
  aleaTable->nbLignes = nbLignes;
  aleaTable->nbColonnes = nbColonnes;
  aleaTable->sommesLignes = (int *)malloc(sizeof(int)*nbLignes);
  aleaTable->sommesColonnes = (int *)malloc(sizeof(int)*nbColonnes);
  
  for (i = 0; i < nbLignes; i++) {
    aleaTable->sommesLignes[i] = sommesLignes[i];
  }
  for (i = 0; i < nbColonnes; i++) {
    aleaTable->sommesColonnes[i] = sommesColonnes[i];
  }
  
  /*on verifie que les deux sommes sont bien egales*/
  int somme1, somme2;
  somme1 = 0;
  for (i = 0; i < aleaTable->nbLignes; i++) {
    somme1 += aleaTable->sommesLignes[i];
  }
  somme2 = 0;
  for (i = 0; i < aleaTable->nbColonnes; i++) {
    somme2 += aleaTable->sommesColonnes[i];
  }
  if (somme1 != somme2) {
    printf("Table de contingence mal definie\n");
    exit(0);
  }
  
  aleaTable->valeurs = (int **)malloc(sizeof(int *)*nbLignes);
  for (i = 0; i < nbLignes; i++) {
    aleaTable->valeurs[i] = (int *)malloc(sizeof(int)*nbColonnes);
  }
}

/***************************************************************************/
/*                                                                         */
/*Procedure verifiant qu'une table de contingence n'a pas d'entree negative*/
/*                                                                         */
/***************************************************************************/

int validite(aleaTable_t aleaTable) {
  int i, j;
  
  if (aleaTable.nbLignes == 2) {
    return 1;
  } else {
    for (i=0;i<aleaTable.nbLignes;i++) {
      for (j=0;j<aleaTable.nbColonnes;j++) {
	if (aleaTable.valeurs[i][j]<0) {
	  return 0;
	}
      }
    }
    return 1;
  }
}

/**********************************************/
/*                                            */
/*Procedure affichant une table de contingence*/
/*                                            */
/**********************************************/

void affiche(aleaTable_t aleaTable) {
  int i, j;
  for (i = 0; i <aleaTable.nbLignes; i++) {
    for (j = 0; j <aleaTable.nbColonnes; j++) {
      printf("%d ",aleaTable.valeurs[i][j]);
    }
    printf("\n");
  }
}

/**********************************************/
/*                                            */
/*Procedure affichant une table de contingence*/
/*                                            */
/**********************************************/

void afficheLineairement(aleaTable_t aleaTable) {
  int i, j;
  for (i = 0; i < aleaTable.nbLignes; i++) {
    for (j = 0; j < aleaTable.nbColonnes; j++) {
      printf(" %d",aleaTable.valeurs[i][j]);
    }
  }
  printf("\n");
}

/**********************************/
/*                                */
/*Procedure calculant la table Max*/
/*                                */
/**********************************/

aleaTable_t calculeTableMax(aleaTable_t *aleaTableEntree) {
  int i, j, k;
  int sommeLigneCourante, *sommesPartiellesColonnes;
  aleaTable_t table;
  
  aleaTable1(aleaTableEntree->nbLignes,aleaTableEntree->nbColonnes,&table);
  sommesPartiellesColonnes = (int *)malloc(sizeof(int)*aleaTableEntree->nbColonnes);
  
  /*traitement a part de la premiere ligne*/
  if (aleaTableEntree->sommesColonnes[0] < aleaTableEntree->sommesLignes[0]) {
    table.valeurs[0][0] = aleaTableEntree->sommesColonnes[0];
  } else {
    table.valeurs[0][0] = aleaTableEntree->sommesLignes[0];
  }
  sommeLigneCourante = table.valeurs[0][0];
  
  for (j = 1; j < aleaTableEntree->nbColonnes; j++) {
    if (aleaTableEntree->sommesColonnes[j] <= aleaTableEntree->sommesLignes[0] - sommeLigneCourante) {
      table.valeurs[0][j] = aleaTableEntree->sommesColonnes[j];
      sommeLigneCourante += table.valeurs[0][j];
    } else {
      table.valeurs[0][j] = aleaTableEntree->sommesLignes[0] - sommeLigneCourante;
      break;
    }
  }
  for (k = j + 1; k < aleaTableEntree->nbColonnes; k++) {
    table.valeurs[0][k] = 0;
  }
  for (i = 0; i < aleaTableEntree->nbColonnes; i++) {
    sommesPartiellesColonnes[i] = table.valeurs[0][i];
  }
  
  /*traitement des lignes suivantes*/
  for (i = 1; i < aleaTableEntree->nbLignes; i++) {
    sommeLigneCourante = 0;
    for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
      if (aleaTableEntree->sommesLignes[i] - sommeLigneCourante >= 
	  aleaTableEntree->sommesColonnes[j] - sommesPartiellesColonnes[j]) {
	table.valeurs[i][j] = aleaTableEntree->sommesColonnes[j] - sommesPartiellesColonnes[j];
	sommeLigneCourante += table.valeurs[i][j];
	sommesPartiellesColonnes[j] = aleaTableEntree->sommesColonnes[j];
      } else {
	table.valeurs[i][j] = aleaTableEntree->sommesLignes[i] - sommeLigneCourante;
	sommesPartiellesColonnes[j] += table.valeurs[i][j];
	break;
      }
    }
    for (k = j + 1; k < aleaTableEntree->nbColonnes; k++) {
      table.valeurs[i][k] = 0;
    }
  }
  
  return table;
}

/**********************************/
/*                                */
/*Procedure calculant la table Min*/
/*                                */
/**********************************/

aleaTable_t calculeTableMin(aleaTable_t *aleaTableEntree) {
  int i, j, k;
  int sommeLigneCourante,*sommesPartiellesColonnes;
  aleaTable_t table;
  
  aleaTable1(aleaTableEntree->nbLignes,aleaTableEntree->nbColonnes,&table);
  sommesPartiellesColonnes = (int *)malloc(sizeof(int)*aleaTableEntree->nbColonnes);
  
  /*traitement a part de la premiere ligne*/
  if (aleaTableEntree->sommesColonnes[aleaTableEntree->nbColonnes - 1] < aleaTableEntree->sommesLignes[0]) {
    table.valeurs[0][aleaTableEntree->nbColonnes - 1] = aleaTableEntree->sommesColonnes[aleaTableEntree->nbColonnes - 1];
  } else {
    table.valeurs[0][aleaTableEntree->nbColonnes - 1] = aleaTableEntree->sommesLignes[0];
  }
  sommeLigneCourante = table.valeurs[0][aleaTableEntree->nbColonnes - 1];
  
  for (j = aleaTableEntree->nbColonnes - 2; j >= 0; j--) {
    if (aleaTableEntree->sommesLignes[0] - sommeLigneCourante >= aleaTableEntree->sommesColonnes[j]) {
      table.valeurs[0][j] = aleaTableEntree->sommesColonnes[j];
      sommeLigneCourante += table.valeurs[0][j];
    } else {
      table.valeurs[0][j] = aleaTableEntree->sommesLignes[0] - sommeLigneCourante;
      break;
    }
  }
  for (k = j - 1; k >= 0; k--) {
    table.valeurs[0][k] = 0;
  }
  for (i = 0; i < aleaTableEntree->nbColonnes; i++) {
    sommesPartiellesColonnes[i] = table.valeurs[0][i];
  }
  
  /*traitement des lignes suivantes*/
  for (i = 1; i < aleaTableEntree->nbLignes; i++) {
    sommeLigneCourante = 0;
    for (j = aleaTableEntree->nbColonnes - 1; j >= 0; j--) {
      if (aleaTableEntree->sommesLignes[i] - sommeLigneCourante >= 
	  aleaTableEntree->sommesColonnes[j] - sommesPartiellesColonnes[j]) {
	table.valeurs[i][j] = aleaTableEntree->sommesColonnes[j] - sommesPartiellesColonnes[j];
	sommeLigneCourante += table.valeurs[i][j];
	sommesPartiellesColonnes[j] = aleaTableEntree->sommesColonnes[j];
	
      } else {
	table.valeurs[i][j] = aleaTableEntree->sommesLignes[i] - sommeLigneCourante;
	sommesPartiellesColonnes[j] += table.valeurs[i][j];
	break;
      }
    }
    for (k = j - 1; k >= 0; k--) {
      table.valeurs[i][k] = 0;
    }
  }
  return table;
}

/**********************************/
/*                                */
/*Procedure calculant la table Max*/
/*                                */
/**********************************/

int egalite(aleaTable_t table1,aleaTable_t table2) {
  int i, j;
  
  for (i = 0; i < table1.nbLignes; i++) {
    for (j = 0; j < table2.nbColonnes; j++) {
      if (table1.valeurs[i][j] != table2.valeurs[i][j]) {
	return 0;
      }
    }
  }
  return 1;
}

/**********************************/
/*                                */
/*Procedure calculant la table Max*/
/*                                */
/**********************************/

void copieTable(aleaTable_t table1,aleaTable_t *table2) {
  int i, j;
  
  table2->nbLignes=table1.nbLignes;
  table2->nbColonnes=table1.nbColonnes;
  table2->valeurs=(int **)malloc(sizeof(int *)*table1.nbLignes);
  for(i=0;i<table1.nbLignes;i++) {
    table2->valeurs[i]=(int *)malloc(sizeof(int)*table1.nbColonnes);
  }
  
  for (i = 0; i < table1.nbLignes; i++) {
    for (j = 0; j < table1.nbColonnes; j++) {
      table2->valeurs[i][j] = table1.valeurs[i][j];
    }
  }
}

/*********************************************************************/
/*                                                                   */
/*Procedure calculant une table avec l'algorithme de Matsui et Kijima*/
/*                                                                   */
/*********************************************************************/

void prochaineTableA2lignes(aleaTable_t *aleaTable,int choixColonne,double lambda) {
  int i, ax, bx, sj, sjp1, theta;
  
  /*Pour deux lignes on applique l'algorithme de Matsui et Kijima*/
  ax = aleaTable->valeurs[0][choixColonne] + aleaTable->valeurs[0][choixColonne + 1];
  bx = aleaTable->valeurs[1][choixColonne] + aleaTable->valeurs[1][choixColonne + 1];
  sj = aleaTable->valeurs[0][choixColonne] + aleaTable->valeurs[1][choixColonne];
  sjp1 = aleaTable->valeurs[0][choixColonne + 1] + aleaTable->valeurs[1][choixColonne + 1];
  
  theta = ax;
  if (bx < theta) {
    theta = bx;
  }
  if (sj < theta) {
    theta = sj;
  }
  if (sjp1 < theta) {
    theta = sjp1;
  }
  theta++;

  
  int min;
  
  if(ax<sj) {
    min=ax;
  } else {
    min=sj;
  }
  aleaTable->valeurs[0][choixColonne] = min - (int)floor(lambda * (double) theta);
  aleaTable->valeurs[0][choixColonne + 1] = ax - aleaTable->valeurs[0][choixColonne];
  aleaTable->valeurs[1][choixColonne] = sj - aleaTable->valeurs[0][choixColonne];
  aleaTable->valeurs[1][choixColonne + 1] = sjp1 - aleaTable->valeurs[0][choixColonne + 1];
}

/**************************************************************************************/
/*                                                                                    */
/*Algorithme de Propp Wilson pour generer de maniere uniforme une table de contingence*/
/*                                                                                    */
/**************************************************************************************/

aleaTable_t proppWilson(aleaTable_t *aleaTableEntree) {
  int i,j,choixColonne,choixDeplacement;
  aleaTable_t tableInf,tableSup;
  
  if (aleaTableEntree->nbLignes == 2) {
    int nbIterationsMax=100000;
    int *choixColonnes;
    double *lambdas;
    
    choixColonnes=(int *)malloc(sizeof(int)*nbIterationsMax);
    lambdas=(double *)malloc(sizeof(double)*nbIterationsMax);

    for(i=0;i<nbIterationsMax;i++) {
      /*on genere choixColonne aleatoirement entre 0 et nbColonnes-1*/
      choixColonnes[i]=(int)floor((double)rand()/(double)RAND_MAX*
				  (double)(aleaTableEntree->nbColonnes-1));  
      lambdas[i]=(double)rand()/(double)RAND_MAX;
    }
 
    /*on cherche d'abord le Max et le Min*/
    tableSup=calculeTableMax(aleaTableEntree);
    tableInf=calculeTableMin(aleaTableEntree);
   
    aleaTable_t tableInf2,tableSup2;
    int tempDepart=nbIterationsMax-1;
    while (1) {
      //printf("tempDepart : %d\n",tempDepart);
      /*initialisation des tables de depart*/
      copieTable(tableSup,&tableSup2);
      copieTable(tableInf,&tableInf2);
      for(i=tempDepart;i<nbIterationsMax;i++) {
	prochaineTableA2lignes(&tableSup2,choixColonnes[i],lambdas[i]);
	prochaineTableA2lignes(&tableInf2,choixColonnes[i],lambdas[i]);
      }
      if(egalite(tableInf2,tableSup2)) {
	break;
      } else {
	tempDepart--;
      }
    }
    return tableSup2;
  } else {
    aleaTable_t table1, table2;
    int *sommesLignesTable1, *sommesColonnesTable1, *sommesLignesTable2, *sommesColonnesTable2;
    
    sommesLignesTable1 = (int *)malloc(sizeof(int)*(aleaTableEntree->nbLignes-1));
    sommesColonnesTable1 = (int *)malloc(sizeof(int)*(aleaTableEntree->nbColonnes));
    sommesLignesTable2 = (int *)malloc(sizeof(int)*(aleaTableEntree->nbLignes-1));
    sommesColonnesTable2 = (int *)malloc(sizeof(int)*(aleaTableEntree->nbColonnes));
    
    /*table1 pour la transition 1*/
    sommesLignesTable1[0] = aleaTableEntree->sommesLignes[0] + aleaTableEntree->sommesLignes[1];
    for (i=1;i<aleaTableEntree->nbLignes-1;i++) {
      sommesLignesTable1[i] = aleaTableEntree->sommesLignes[i+1];
    }
    for (i=0;i<aleaTableEntree->nbColonnes;i++) {
      sommesColonnesTable1[i] = aleaTableEntree->sommesColonnes[i];
    }
    
    /*table2 pour la transition 2*/
    sommesLignesTable2[aleaTableEntree->nbLignes - 2] = 
      aleaTableEntree->sommesLignes[aleaTableEntree->nbLignes - 2] + 
      aleaTableEntree->sommesLignes[aleaTableEntree->nbLignes - 1];
    for (i = 0; i < aleaTableEntree->nbLignes - 2; i++) {
      sommesLignesTable2[i] = aleaTableEntree->sommesLignes[i];
    }
    for (i = 0; i < aleaTableEntree->nbColonnes; i++) {
      sommesColonnesTable2[i] = aleaTableEntree->sommesColonnes[i];
    }
    
    while (1) {
      tableSup=calculeTableMax(aleaTableEntree);

      aleaTable2(aleaTableEntree->nbLignes-1,aleaTableEntree->nbColonnes, 
		 sommesLignesTable1,sommesColonnesTable1,&table1);
      table1=proppWilson(&table1);
      aleaTable2(aleaTableEntree->nbLignes-1,aleaTableEntree->nbColonnes, 
		 sommesLignesTable2,sommesColonnesTable2,&table2);
      table2=proppWilson(&table2);
      
      if ((double)rand()/(double)RAND_MAX < 0.5) {
	/*transition 1*/
	for (i = 2; i < aleaTableEntree->nbLignes; i++) {
	  for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	    tableSup.valeurs[i][j] = table1.valeurs[i - 1][j];
	  }
	}
	for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	  tableSup.valeurs[1][j] = table1.valeurs[0][j] - tableSup.valeurs[0][j];
	}
	
	/*puis transition 2*/
	for (i = 0; i < aleaTableEntree->nbLignes - 2; i++) {
	  for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	    tableSup.valeurs[i][j] = table2.valeurs[i][j];
	  }
	}
	for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	  tableSup.valeurs[aleaTableEntree->nbLignes - 2][j] = 
	    table2.valeurs[aleaTableEntree->nbLignes - 2][j] - 
	    tableSup.valeurs[aleaTableEntree->nbLignes - 1][j];
	}
	
      } else {
	/*transition 2*/
	for (i = 0; i < aleaTableEntree->nbLignes - 2; i++) {
	  for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	    tableSup.valeurs[i][j] = table2.valeurs[i][j];
	  }
	}
	for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	  tableSup.valeurs[aleaTableEntree->nbLignes - 2][j] = 
	    table2.valeurs[aleaTableEntree->nbLignes - 2][j] - 
	    tableSup.valeurs[aleaTableEntree->nbLignes - 1][j];
	}
	/*puis transition 1*/
	for (i = 2; i < aleaTableEntree->nbLignes; i++) {
	  for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	    tableSup.valeurs[i][j] = table1.valeurs[i - 1][j];
	  }
	}
	for (j = 0; j < aleaTableEntree->nbColonnes; j++) {
	  tableSup.valeurs[1][j] = table1.valeurs[0][j] - tableSup.valeurs[0][j];
	}
      }
      if (validite(tableSup)) {
	break;
      } else {
	//printf("X\n");
      }
    }
    return tableSup;
  }
}

/********************************************************************/
/*                                                                  */
/*Programme de generation de tables de contingence de maniere exacte*/
/*                                                                  */
/********************************************************************/

int main(int argc,char **argv) {
  srand(getpid());
  
  aleaTable_t tableEntree,tableGeneree;

  if(argc<2) {
    printf("Usage : aleaTable file\n");
    exit(0);
  }
  lectureFichier(argv[1],&tableEntree);
  tableGeneree=proppWilson(&tableEntree);
  affiche(tableGeneree);
  exit(0);
}

